/*
 * Sun RPC is a product of Sun Microsystems, Inc. and is provided for
 * unrestricted use provided that this legend is included on all tape
 * media and as a part of the software program in whole or part.  Users
 * may copy or modify Sun RPC without charge, but are not authorized
 * to license or distribute it to anyone else except as part of a product or
 * program developed by the user.
 * 
 * SUN RPC IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING THE
 * WARRANTIES OF DESIGN, MERCHANTIBILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 * 
 * Sun RPC is provided with no support and without any obligation on the
 * part of Sun Microsystems, Inc. to assist in its use, correction,
 * modification or enhancement.
 * 
 * SUN MICROSYSTEMS, INC. SHALL HAVE NO LIABILITY WITH RESPECT TO THE
 * INFRINGEMENT OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY SUN RPC
 * OR ANY PART THEREOF.
 * 
 * In no event will Sun Microsystems, Inc. be liable for any lost revenue
 * or profits or other special, indirect and consequential damages, even if
 * Sun has been advised of the possibility of such damages.
 * 
 * Sun Microsystems, Inc.
 * 2550 Garcia Avenue
 * Mountain View, California  94043
 */

#if !defined(lint) && defined(SCCSIDS)
static char sccsid[] = "@(#)test.c  1.3  92/06/22  Copyr 1988 Sun Micro";
#endif

/*
 * This file contains the actual test routines for the various test
 * cases for thruput/latency/binding performance tests.
 */

#include <stdio.h>
#include <rpc/rpc.h>
#include <sys/fcntl.h>
#include "test.h"

#define	MILLION 1000000.0
#define	KBPERSEC(x) (x * 1024.0)/MILLION

extern long diff_time();
extern void add_time();
extern char *print_time();
extern float mspercall();
extern void fatal();


/* global stuff */
char host[32];		/* name of host the server is running on */
char transport[32];	/* transport type - datagram_n, curcuit_n, udp/tcp */
CLIENT *clnt;		/* clnt handle - created in  client.c */
struct netconfig *nconf;/* netconfig pointer for transport */

/* global - should be static if client.c does not want to use this */
bool_t latTest;		/* latency test - NULL call */
bool_t bindTest;	/* binding test */
bool_t thruTest;	/* thruput test */
bool_t timeEachCall;	/* give timing for each separate call */
bool_t onePktTest;	/* make calls with 1400 bytes */
bool_t intArrTest;	/* make calls with integer array */
bool_t relaxedPrint;	/* print the numbers in the end */
bool_t statCalculate;	/* calculate statistics */
bool_t Verbose;		/* print detailed info about the test being run */
int calls;		/* number of calls to make */
u_int arglen;		/* argument length for int. array, opaque data */
u_int bytes;		/* Number of bytes in each RPC */

static void usage();

void
parseArgs(argc, argv)
	int argc;
	char *argv[];
{
	char c;
	extern char *optarg;	/* command-line option string */
	int cmds = 0;

	/* defaults */
	gethostname(host, 32);
	strcpy(transport, "datagram_n");
	calls = 1000;
	bindTest = FALSE;
	latTest = FALSE;
	thruTest = FALSE;
	timeEachCall = FALSE;
	onePktTest = FALSE;
	intArrTest = FALSE;
	relaxedPrint = FALSE;
	statCalculate = FALSE;
	arglen = 1400;

	/* get command line options */
	while ((c = getopt(argc, argv, "beh:n:s:t:BEINORSTXLV")) != EOF) {
		switch (c) {
		case 'h':
			strcpy(host, optarg);
			break;
		case 'n':
			calls = atoi(optarg);
			break;
		case 's':
			arglen = atoi(optarg);
			break;
		case 't':
			strcpy(transport, optarg);
			break;
		case 'B':
			bindTest = TRUE;
			cmds++;
			break;
		case 'E':
			timeEachCall = TRUE;
			break;
		case 'I':
			intArrTest = TRUE;
			cmds++;
			break;
		case 'N':
			latTest = TRUE;
			cmds++;
			break;
		case 'O':
			onePktTest = TRUE;
			cmds++;
			break;
		case 'R':
			relaxedPrint = TRUE;
			/* fall through */
		case 'S':
			timeEachCall = TRUE;
			statCalculate = TRUE;
			break;
		case 'T':
			thruTest = TRUE;
			cmds++;
			break;
		case 'V':
			Verbose = TRUE;
			break;
		case '?':	/* error */
			usage();
			break;
		}
	}

	if (cmds == 0) {
		fprintf(stderr, "Please specify type of test\n");
		usage();
	} else if (cmds > 1) {
		fprintf(stderr, "Only one type of test can be specified\n");
		usage();
	}
	if ((bindTest == TRUE) && (strcmp(transport, "raw") == 0)) {
		fprintf(stderr, "Bind test not allowed with raw transport\n");
		exit(1);
	}
}

void
runTests()
{

#ifdef MICROTIMER
	long before, after, total;
#else
	struct timeval before, after, total;
#endif

	long diff_usecs;
	u_long secs, usecs;
	int i;
	float div, thruput;
	/* args */
	opaqueArgVar oArgVar;
	opaqueArg1400 arg1400;
	UintArgVar intarg;
	int	integer_cksum = 0;
	int	*res;
	struct rpc_err err;

#ifdef MICROTIMER
	total = 0;
#else
	total.tv_sec = total.tv_usec = 0;
#endif

	if (Verbose) {
	    printf("\nTransport type is %s", transport);
#ifdef SVR4
		if (nconf == NULL) {
			printf(".\n");
		} else {
			printf(",  is %s.\n", nconf->nc_netid);
		}
#else
	    printf(".\n");
#endif
	}
	oArgVar.data.data_len = arglen;
	if ((oArgVar.data.data_val = (char *) malloc(arglen))
		== NULL) {
		fatal("failed in malloc");
	}
	intarg.UintArgVar_len = arglen;
	if ((intarg.UintArgVar_val = (int *) malloc(sizeof (int) * arglen))
		== NULL) {
		fatal("failed in malloc");
	} else {
		for (i = 0; i < arglen; i++) {
			intarg.UintArgVar_val[i] = i;
#ifdef DO_CHECKSUM
			integer_cksum += i;
#endif
		}
	}
	if (!timeEachCall) {
	    if (Verbose) {
		    printf("%d calls will be timed together.\n", calls);
	    }
	    if (latTest) {

		if (Verbose) {
			printf("Latency test: timing NULL RPC calls.\n\n");
		}
		getmicrotime(&before, NULL);
		for (i = 0; i < calls; i++) {
			res = null_call_1(&i, clnt);
			if (res == NULL) {
				clnt_geterr(clnt, &err);
				printf("RPC error on call %d: %s\n", i,
					clnt_sperrno(err.re_status));
				break;
			} else if (*res != i) {
    printf("Did not get correct results! Sent: %d, got back %d\n", i, *res);
			}
		}
		getmicrotime(&after, NULL);
		if (res == NULL) {
			before = after;
		}
		diff_usecs = diff_time(&before, &after);
		printf("for %d NULL calls: %s time/call: %.2f ms\n",
			calls, print_time(&before, &after),
			mspercall(diff_usecs, calls));
	    }
	    if (thruTest) {
		if (Verbose) {
		    printf(
		     "Throughput test: timing %d byte (opaque) RPC calls.\n\n",
		      arglen);
		}

		if (check_transport(clnt, arglen)) {	/* xprt okay */

			getmicrotime(&before, NULL);
			for (i = 0; i < calls; i++) {
				res = thru_o_var_1(&oArgVar, clnt);
				if (res == NULL) {
					clnt_geterr(clnt, &err);
					printf("RPC error on call %d: %s\n", i,
						clnt_sperrno(err.re_status));
					break;
				}
			}
			getmicrotime(&after, NULL);
			if (res == NULL) {
				before = after;
			}
			bytes = (calls * (arglen + 4));	/* 4 for int in len */
			diff_usecs = diff_time(&before, &after);
			div = KBPERSEC(diff_usecs);
			if (div)
				thruput = (bytes * 1.0) / div;
			else
				thruput = 0.0;

			printf("for %d opaqueVarArg <%d>calls:", calls,
				arglen);
			printf(
			" %d bytes thruput: %.2f Kb/s %s time/call: %.2f ms\n",
				bytes, thruput, print_time(&before, &after),
				mspercall(diff_usecs, calls));
		}
	    }
	    if (onePktTest) {
		if (Verbose) {
		    printf(
	"Single-packet test: timing 1400 byte (opaque) RPC calls.\n\n");
		}
		getmicrotime(&before, NULL);
		for (i = 0; i < calls; i++) {
			res = thru_o_1400_1(&arg1400, clnt);
			if (res == NULL) {
				clnt_geterr(clnt, &err);
				printf("RPC error on call %d: %s\n", i,
					clnt_sperrno(err.re_status));
				break;
			}
		}
		getmicrotime(&after, NULL);
		if (res == NULL) {
			before = after;
		}
		diff_usecs = diff_time(&before, &after);
		bytes = (calls * 1400);
		div = KBPERSEC(diff_usecs);
		if (div)
			thruput = (bytes * 1.0) / div;
		else
			thruput = 0.0;

		printf("for %d opaque1400Arg calls:", calls);
		printf(" %d bytes %.2f Kb/s %s time/call: %.2f ms\n",
			bytes, thruput, print_time(&before, &after),
			mspercall(diff_usecs, calls));
	    }
	    if (intArrTest) {
		int *(*local)();

		if (Verbose) {
		    printf(
"Integer array test: %d element integer array (%d bytes) RPC calls.\n\n",
		      arglen, (arglen + 1) * sizeof (int));
		}

#ifdef SVR4
		if (nconf != NULL && nconf->nc_semantics != NC_TPI_CLTS)
#else
		if (strcmp(transport, "tcp") == 0)
#endif
			local = (int *(*)()) thru_ti_var_1;
		else /* for raw and datagram */
			local = (int *(*)()) thru_ui_var_1;


		if (check_transport(clnt, arglen*sizeof (int))) {

			thru_i_reset_1(&res, clnt);
			res = (*local)(&intarg, clnt);	/* warm up checksum */
			if (res == NULL) {
				clnt_geterr(clnt, &err);
				printf("RPC error on FIRST call: %s\n",
					clnt_sperrno(err.re_status));
			}
			getmicrotime(&before, NULL);
			for (i = 0; i < calls; i++) {
				res = (*local)(&intarg, clnt);
				if (res == NULL) {
					clnt_geterr(clnt, &err);
					printf("RPC error on call %d: %s\n", i,
						clnt_sperrno(err.re_status));
					break;
				}
#ifdef DO_CHECKSUM
				else if (*res != integer_cksum) {
		printf("Checksum on call %d was wrong! Should be %d, was %d\n",
					i, integer_cksum, *res);
				}
#endif
			}
			getmicrotime(&after, NULL);
			if (res == NULL) {
				before = after;
			}
			diff_usecs = diff_time(&before, &after);
			/* 4 for len of array */
			bytes = (calls * (arglen * sizeof (int) + 4));
			div = KBPERSEC(diff_usecs);
			if (div)
				thruput = (bytes * 1.0) / div;
			else
				thruput = 0.0;
			printf("for %d intVarArg <%d> calls:", calls, arglen);
			printf(" %d bytes %.2f Kb/s %s time/call: %.2f ms\n",
				bytes, thruput, print_time(&before, &after),
				mspercall(diff_usecs, calls));
		    }
	    }
	    if (bindTest) {
		if (Verbose) {
		    printf(
			"Client handle creation test: for host %s.\n\n", host);
		}
		getmicrotime(&before, NULL);
		for (i = 0; i < calls; i++) {
			clnt = clnt_create(host, PERF, VERS_ONE, transport);

			if (clnt)
				clnt_destroy(clnt);
			else {
				clnt_pcreateerror(host);
				break;
			}
		}
		getmicrotime(&after, NULL);
		if (clnt == NULL) {
			before = after;
		}
		diff_usecs = diff_time(&before, &after);
		printf("time for %d bind (clnt_create & destroy) calls:",
			calls);
		printf(" %s time/call: %.2f ms\n",
			print_time(&before, &after),
			mspercall(diff_usecs, calls));
	    }
	} else {	/* time each call */
	    if (statCalculate) {
		init_data(calls);
	    }
	    if (Verbose) {
		    printf("Each of %d calls will be timed separately.\n",
			calls);
	    }
	    if (latTest) {

		if (Verbose) {
			printf("Latency test: timing NULL RPC calls.\n\n");
		}
		for (i = 1; i <= calls; i++) {
			getmicrotime(&before, NULL);
			res = null_call_1(&i, clnt);
			getmicrotime(&after, NULL);
			if (res == NULL) {
				clnt_geterr(clnt, &err);
				printf("RPC error on call %d: %s\n", i,
					clnt_sperrno(err.re_status));
				before = after;
			} else if (*res != i) {
    printf("Did not get correct results! Sent: %d, got back %d\n", i, *res);
			}
			if (res == NULL) {
				before = after;
			}
			add_time(&before, &after, &total);
			if (statCalculate)
				add_data(diff_time(&before, &after));
			if (relaxedPrint == FALSE)
				printf("NULL call(%3d): %s\n", i,
					print_time(&before, &after));
		}
	    }
	    if (thruTest) {
		if (Verbose) {
		    printf(
		     "Throughput test: timing %d byte (opaque) RPC calls.\n\n",
		      arglen);
		}
		if (check_transport(clnt, arglen)) {	/* xprt okay */
		    bytes = arglen + 4;	/* 4 for int in len */
		    for (i = 1; i <= calls; i++) {
			getmicrotime(&before, NULL);
			res = thru_o_var_1(&oArgVar, clnt);
			getmicrotime(&after, NULL);
			if (res == NULL) {
				clnt_geterr(clnt, &err);
				printf("RPC error on call %d: %s\n", i,
					clnt_sperrno(err.re_status));
				before = after;
			}
			add_time(&before, &after, &total);
			if (statCalculate)
				add_data(diff_time(&before, &after));
			if (relaxedPrint == FALSE)
			    printf("opaqueVarArg (%3d) call, %d bytes: %s\n",
				i, arglen, print_time(&before, &after));
		    }
		}
	    }
	    if (intArrTest) {
		struct rpc_err err;
		int *(*local)();

		if (Verbose) {
		    printf(
"Integer array test: %d element integer array (%d bytes) RPC calls.\n\n",
		      arglen, (arglen + 1) * sizeof (int));
		}
#ifdef SVR4
		if (nconf != NULL && nconf->nc_semantics != NC_TPI_CLTS)
#else
		if (strcmp(transport, "tcp") == 0)
#endif
			local = (int *(*)()) thru_ti_var_1;
		else
			local = (int *(*)()) thru_ui_var_1;


		bytes = (arglen + 1) * sizeof (int);
		if (check_transport(clnt, bytes)) {	/* xprt okay */

		    thru_i_reset_1(&res, clnt);
		    res = (*local)(&intarg, clnt);	/* warm up checksum */
		    if (res == NULL) {
			clnt_geterr(clnt, &err);
			printf("RPC error on FIRST call: %s\n",
				clnt_sperrno(err.re_status));
		    }
		    for (i = 1; i <= calls; i++) {
			getmicrotime(&before, NULL);
			res = (*local)(&intarg, clnt);
			getmicrotime(&after, NULL);
			if (res == NULL) {
				clnt_geterr(clnt, &err);
				printf("RPC error on call %d: %s\n", i,
					clnt_sperrno(err.re_status));
				before = after;
			}
#ifdef DO_CHECKSUM
			else if (*res != integer_cksum) {
	printf("Checksum on call %d was wrong! Should be %d, was %d\n",
				i, integer_cksum, *res);
			}
#endif
			add_time(&before, &after, &total);
			if (statCalculate)
				add_data(diff_time(&before, &after));
			if (relaxedPrint == FALSE)
				printf("intVarArg call(%3d): %d integers %s\n",
				       i, arglen, print_time(&before, &after));
		    }
		}
	    }
	    if (onePktTest) {
		if (Verbose) {
		    printf(
			"Single-packet test: timing 1400 byte RPC calls.\n\n");
		}
		bytes = 1400;
		for (i = 1; i <= calls; i++) {
			getmicrotime(&before, NULL);
			res = thru_o_1400_1(&arg1400, clnt);
			getmicrotime(&after, NULL);
			if (res == NULL) {
				clnt_geterr(clnt, &err);
				printf("RPC error on call %d: %s\n", i,
					clnt_sperrno(err.re_status));
				before = after;
			}
			add_time(&before, &after, &total);
			if (statCalculate)
				add_data(diff_time(&before, &after));
			if (relaxedPrint == FALSE)
				printf("onePkt call(%3d): %d bytes %s\n",
					i, bytes, print_time(&before, &after));
		}
	    }
	    if (bindTest) {
		if (Verbose) {
		    printf(
			"Client handle creation test: for host %s.\n\n", host);
		}
		for (i = 0; i < calls; i++) {
			getmicrotime(&before, NULL);
			clnt = clnt_create(host, PERF, VERS_ONE, transport);
			/*
			 * Since our typical interest is in time required to
			 * create a client, measure only that for binding time
			 */
			getmicrotime (&after, NULL);
			if (clnt) {
				add_time(&before, &after, &total);
				clnt_destroy(clnt);
			}
			else {
				clnt_pcreateerror(host);
				before = after;
			}
			if (statCalculate)
				add_data(diff_time(&before, &after));
			if (relaxedPrint == FALSE)
				printf("bind call(%3d): %s\n",
					i, print_time(&before, &after));
		}
	    }
#ifndef MICROTIMER
	    printf("total time for %d calls = %ld secs %ld usecs\n",
		calls, total.tv_sec, total.tv_usec);
#else
	    usecs = total;
	    if (usecs > 1000000)
		secs = usecs / 1000000;
	    else
		secs = 0;
	    usecs = usecs - secs * 1000000;
	    printf("total time for %d calls = %ld secs %ld usecs\n",
		calls, secs, usecs);
#endif

	}   /* Else */
	if (statCalculate) {
		if (relaxedPrint == TRUE)
			print_all_data();
		print_real_ave();
	}
	printf("\n");
}

static void
usage()
{
	fprintf(stderr, "Usage:\n\
	-B <run binding test - default is performance>\n\
	-E <time each call - default is FALSE>\n\
	-I <run int array test - size of array in arglen (-s)>\n\
	-N <run latency - NULL call test>\n\
	-O <make calls for 1 Pkt - 1400 bytes>\n\
	-R <relaxed print. times separately, does stats and prints at the end>\n\
	-S <calculates statistics. assumes -E>\n\
	-T <run thruput test - opaque data - size in (-s)>\n\
	-h [server_host - defaults to local]\n\
	-n [number_of_calls - defaults to 1000]\n\
	-s [size_of_argument - defaults to 1400]\n\
	-t [transport - defaults to datagram_n]\n");
	exit(1);
}

#define	RPC_RQST_SZ 28		/* size of an RPC header with AUTH_NULL */


#ifdef SVR4
int
check_transport(clnt, size)
CLIENT *clnt;
int size;
{
	struct t_info tfo;
	int fd;

	if (nconf == NULL) {
		return (1);		/* probably the raw tests */
	}

	fd = t_open(nconf->nc_device, O_RDWR, &tfo);
	if (fd == -1) {
		fprintf(stderr, "Transport device %s cannot be opened.\n",
			nconf->nc_device);
		t_error(nconf->nc_netid);
		return (0);
	}
	if (tfo.tsdu > 0 && tfo.tsdu < size + RPC_RQST_SZ) {
		fprintf(stderr,
	"Transport device %s cannot send requested amount of data (%d).\n",
		nconf->nc_device, size + RPC_RQST_SZ);
		fprintf(stderr,	"Maximum segment size is %d\n", tfo.tsdu);
		return (0);
	}
	return (1);
}
#else
int
check_transport(clnt, size)
CLIENT *clnt;
int size;
{
	return (1);
}
#endif
